from decimal import Decimal
from functools import wraps
from typing import Any, Callable


def get_type(obj):
    if obj is None:
        return "null"
    if isinstance(obj, str):
        return "string"
    if isinstance(obj, bool):
        return "boolean"
    if isinstance(obj, list):
        return "array"
    if isinstance(obj, dict):
        return "object"
    if isinstance(obj, int | float | Decimal):
        return "number"
    raise TypeError


def when(type: str):
    def inner(func: Callable[[Any], bool]):
        @wraps(func)
        def decorated(obj: str):
            if is_type(obj, type):
                return func(obj)
            return NotImplemented

        decorated.__wrapped__ = func  # type: ignore
        return decorated

    return inner


def is_type(obj, *types: str) -> bool:
    obj_type = get_type(obj)
    print("grre", obj_type, types)
    if obj_type in types:
        return True
    if obj_type == "number" and "integer" in types and is_integer(obj):
        return True
    return False


def is_integer(obj) -> bool:
    return obj == int(obj)


@when("string")
def is_date_time(obj: str) -> bool:
    pass


@when("string")
def is_date(obj: str) -> bool:
    pass


@when("string")
def is_time(obj: str) -> bool:
    pass


@when("string")
def is_duration(obj: str) -> bool:
    pass


@when("string")
def is_email(obj: str) -> bool:
    pass


@when("string")
def is_idn_email(obj: str) -> bool:
    pass


@when("string")
def is_hostname(obj: str) -> bool:
    pass


@when("string")
def is_idn_hostname(obj: str) -> bool:
    pass


@when("string")
def is_ipv4(obj: str) -> bool:
    pass


@when("string")
def is_ipv6(obj: str) -> bool:
    pass


@when("string")
def is_uri(obj: str) -> bool:
    pass


@when("string")
def is_uri_reference(obj: str) -> bool:
    pass


@when("string")
def is_iri(obj: str) -> bool:
    pass


@when("string")
def is_iri_reference(obj: str) -> bool:
    pass


@when("string")
def is_uuid(obj: str) -> bool:
    pass


@when("string")
def is_uri_template(obj: str) -> bool:
    pass


@when("string")
def is_json_pointer(obj: str) -> bool:
    pass


@when("string")
def is_relative_json_pointer(obj: str) -> bool:
    pass


@when("string")
def is_regex(obj: str) -> bool:
    pass



from __future__ import annotations

import json
from pathlib import Path
from urllib.parse import urljoin

import httpx


class Backend:
    def applicable(self, document_uri: str) -> bool:
        ...

    def get(self, document_uri: str) -> tuple[dict, str]:
        ...


class BackendError(Exception):
    pass


class FileSystemBackend(Backend):
    def __init__(self, dirname: Path, prefix: str = ""):
        self.dirname = dirname
        self.prefix = prefix

    def applicable(self, document_uri: str) -> bool:
        return document_uri.startswith(self.prefix)

    def get(self, document_uri: str) -> tuple[dict, str]:
        uri, *_ = document_uri.partition("#")
        path = self.dirname / (uri.removeprefix(self.prefix) + ".json")
        try:
            return json.loads(path.read_text()), uri
        except FileNotFoundError as error:
            raise BackendError(f"unable to load {document_uri}") from error


class HTTPBackend(Backend):
    def __init__(self, base_url: str, prefix: str = ""):
        self.base_url = base_url
        self.prefix = prefix

    def applicable(self, document_uri: str) -> bool:
        return document_uri.startswith(self.prefix)

    def get(self, document_uri: str) -> tuple[dict, str]:
        uri, *_ = document_uri.partition("#")
        url = urljoin(self.base_url, uri.removeprefix(self.prefix))
        try:
            response = httpx.get(url)
            response.raise_for_status()
            return json.loads(response.text), uri
        except httpx.HTTPStatusError as error:
            raise BackendError(f"unable to load {document_uri}") from error



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Generic, Iterable, TypeVar

from .ref import push_fragments

if TYPE_CHECKING:
    from .evaluators import Evaluate
    from .loaders import Loader
    from .res import Document, Schema

T = TypeVar("T")


def registry_factory() -> Registry:
    from ilia.dialects import draft202021

    from .loaders import Loader

    extractor: MultiMethod[Iterable[tuple[dict | bool, str]]] = MultiMethod()
    extractor.register(draft202021.DIALECT, draft202021.extract)

    finalizer: MultiMethod[Document] = MultiMethod()
    finalizer.register(draft202021.DIALECT, draft202021.finalize)

    schema_compiler: MultiMethod[Evaluate] = MultiMethod()
    schema_compiler.register(draft202021.DIALECT, draft202021.compile_schema)

    loader = Loader([])
    return Registry(
        finalizer=finalizer,
        extractor=extractor,
        schema_compiler=schema_compiler,  # type:ignore
        loader=loader,
    )


class Registry:
    formats: dict

    def __init__(
        self,
        finalizer: Callable[[Document], None],
        extractor: Callable[[Document], Iterable[Document]],
        schema_compiler: Callable[[Document], Evaluate],
        loader: Loader = None,
    ):
        self.documents: list[Document] = []
        self.finalizer = finalizer
        self.extractor = extractor
        self.schema_compiler = schema_compiler
        self.formats = {}
        self.loader = loader

    def load(self, uri: str) -> tuple[Schema, str]:
        if self.loader:
            return self.loader.get(uri)
        raise

    def finalize(self, doc: Document):
        return self.finalizer(doc)

    def extract(self, doc: Document):
        return self.extractor(doc)

    def compile_schema(self, doc: Document):
        return self.schema_compiler(doc)

    def find(self, parent: Document, key: str) -> Document | None:
        relative_location = push_fragments(parent.relative_location, key)
        for doc in self.documents:
            if doc.parent is parent and doc.relative_location == relative_location:
                return doc
        return None

    def find_sequence(
        self, parent: Document[dict], key: str
    ) -> list[tuple[int, Document]] | None:
        z: list[Any] | None
        if (z := parent.schema.get(key)) is not None:
            sequence = []
            for index, _ in enumerate(z):
                relative_location = push_fragments(parent.relative_location, key, index)
                for doc in self.documents:
                    if (
                        doc.parent is parent
                        and doc.relative_location == relative_location
                    ):
                        sequence.append((index, doc))
                        break
                else:
                    raise
            return sequence
        return None

    def find_mapping(
        self, parent: Document[dict], key: str
    ) -> list[tuple[str, Document]] | None:
        z: dict[str, Any] | None
        if (z := parent.schema.get(key)) is not None:
            sequence = []
            for index in z:
                relative_location = push_fragments(parent.relative_location, key, index)
                for doc in self.documents:
                    if (
                        doc.parent is parent
                        and doc.relative_location == relative_location
                    ):
                        sequence.append((index, doc))
                        break
                else:
                    raise
            return sequence
        return None


class MultiMethod(Generic[T]):
    def __init__(self):
        self.methods: dict[str, Callable[[Document], T]] = {}

    def __call__(self, doc: Document):
        return self[doc.dialect](doc)

    def __getitem__(self, dialect):
        return self.methods[dialect]

    def register(self, dialect: str, func: Callable[[Document], T] = None):
        def inner(func: Callable[[Document], T]):
            self.methods[dialect] = func
            return func

        if func:
            return inner(func)
        return inner



from __future__ import annotations

from functools import cached_property


class Ref(str):
    @cached_property
    def uri(self):
        uri, *_ = self.partition("#")
        if uri:
            return uri

    @cached_property
    def anchor(self):
        *_, fragment = self.partition("#")
        if fragment and not fragment.startswith("/"):
            return fragment

    @cached_property
    def pointer(self):
        *_, fragment = self.partition("#")
        if fragment and fragment.startswith("/"):
            return fragment

    @cached_property
    def decomposed(self):
        return (self.uri, self.pointer, self.anchor)


def push_fragments(a: str, *fragments: str | int):
    output = a
    for fragment in fragments:
        output = output + "/" + str(fragment)
    return output


def push_pointer(a: str, *fragments: str | int):
    output = a or ""
    if "#" not in output:
        output = output + "#"
    for fragment in fragments:
        output = output + "/" + str(fragment)
    return output



from __future__ import annotations

from collections import deque
from dataclasses import dataclass, field
from decimal import Decimal
from functools import reduce
from types import NotImplementedType
from typing import (
    Any,
    Callable,
    Generic,
    Iterable,
    Iterator,
    Protocol,
    Type,
    TypeVar,
    cast,
    overload,
)

from .checkers import is_type
from .ref import push_fragments

T = TypeVar("T", covariant=True)
U = TypeVar("U")


class Keyword(Protocol):
    @property
    def relative_location(self) -> str:
        ...

    @property
    def absolute_keyword_location(self) -> str:
        ...


@dataclass
class Instance(Generic[T]):
    location: str
    value: T
    parent: Instance | None = None

    @overload
    def __iter__(self: Instance[dict]) -> Iterator[tuple[str, Instance]]:
        ...

    @overload
    def __iter__(self: Instance[list]) -> Iterator[tuple[int, Instance]]:
        ...

    def __iter__(self) -> Iterator[tuple[str | int, Instance]]:
        if isinstance(self.value, dict):
            iterable: Iterable[tuple[str | int, Any]] = self.value.items()
        elif isinstance(self.value, list):
            iterable = enumerate(self.value)
        else:
            raise AttributeError(f"Instance[{type(self.value)}] is not iterable")
        for name, member in iterable:
            yield name, Instance(
                location=push_fragments(self.location, name), value=member, parent=self
            )

    @property
    def names(self) -> InstanceContext[list[str]]:
        if isinstance(self.value, dict):
            names = list(self.value)
            return self.ctx(names)
        else:
            raise AttributeError(f"Instance[{type(self.value)}] is not iterable")

    def ctx(self, value: U) -> InstanceContext[U]:
        return InstanceContext(value=value, instance=self)


@dataclass
class InstanceContext(Generic[T]):
    value: T
    instance: Instance

    @property
    def location(self):
        return self.instance.location

    def __iter__(self) -> Iterator[InstanceContext]:
        if isinstance(self.value, list):
            for element in self.value:
                yield InstanceContext(value=element, instance=self.instance)


class Evaluable(Protocol[T]):
    @property
    def value(self) -> T:
        ...

    @property
    def location(self) -> str:
        ...


@dataclass
class Evaluation:
    valid: bool
    relative_location: str
    absolute_keyword_location: str
    instance: Instance
    value: Any = None
    reason: str | None = None
    evaluations: list[Evaluation] = field(default_factory=list)

    @property
    def instance_value(self):
        return self.instance.value

    @property
    def instance_location(self):
        return self.instance.location

    @property
    def keyword(self):
        return self.relative_location.rpartition("/")[-1]

    def __post_init__(self):
        if self.valid and self.reason is not None:
            raise Exception("reason is for errors")


class Evaluate(Protocol):
    def __call__(self, _0: Evaluable[T], /) -> Iterable[Evaluation]:
        raise NotImplementedError


def render(
    keyword: Keyword,
    subject: Instance | InstanceContext,
    valid: bool,
    value: Any = None,
    reason: str = None,
    evaluations: list[Evaluation] = None,
) -> Evaluation:
    if isinstance(subject, InstanceContext):
        instance = subject.instance
    else:
        instance = subject
    return Evaluation(
        valid=valid,
        value=value if valid else None,
        reason=reason if not valid else None,
        relative_location=keyword.relative_location,
        absolute_keyword_location=keyword.absolute_keyword_location,
        instance=instance,
        evaluations=evaluations or [],
    )


class AlwaysTrue:
    def __init__(self, relative_location: str, absolute_keyword_location: str):
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        yield render(self, subject=instance, valid=True)


class AlwaysFalse:
    def __init__(self, relative_location: str, absolute_keyword_location: str):
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        yield render(self, subject=instance, valid=False)

    def __repr__(self):
        return "AlwaysFalse()"


class Cluster:
    def __init__(
        self,
        evaluators: list[Evaluate],
        unevaluated_items: UnevaluatedItems,
        unevaluated_props: UnevaluatedProperties,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluators = evaluators
        self.unevaluated_items = unevaluated_items
        self.unevaluated_props = unevaluated_props
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        for evaluator in self.evaluators:
            evaluations += evaluator(instance)
        evaluations += self.yield_unevaluated_items(instance, evaluations)
        evaluations += self.yield_unevaluated_props(instance, evaluations)
        valid = not any(e.valid is False for e in evaluations)
        yield render(self, subject=instance, valid=valid, evaluations=evaluations)

    def __repr__(self):
        return f"Cluster(evaluators={self.evaluators!r}, unevaluated_items={self.unevaluated_items!r}, unevaluated_props={self.unevaluated_props!r})"

    def yield_unevaluated_items(
        self, instance: Instance, evaluations: list[Evaluation]
    ):
        if self.unevaluated_items and isinstance(instance.value, list):
            evaluated = get_evaluated_items(instance, evaluations)
            yield from self.unevaluated_items(instance, ignore=evaluated)

    def yield_unevaluated_props(
        self, instance: Instance, evaluations: list[Evaluation]
    ):
        if self.unevaluated_props and isinstance(instance.value, dict):
            evaluated = get_evaluated_props(instance, evaluations)
            yield from self.unevaluated_props(instance, ignore=evaluated)


def get_evaluated_items(
    instance: Instance[list], evaluations: list[Evaluation]
) -> set[int]:
    evaluated: set[int] = set()
    for evaluation in walk_annotations(instance, evaluations):
        if evaluation.keyword in (
            "prefixItems",
            "items",
            "contains",
            "unevaluatedItems",
        ):
            evaluated |= evaluation.value
    return evaluated


def get_evaluated_props(
    instance: Instance[dict], evaluations: list[Evaluation]
) -> set[str]:
    evaluated: set[str] = set()
    for evaluation in walk_annotations(instance, evaluations):
        if evaluation.keyword in (
            "properties",
            "patternProperties",
            "additionalProperties",
            "unevaluatedProperties",
        ):
            evaluated |= evaluation.value
    return evaluated


def walk_annotations(instance: Instance, evaluations: list[Evaluation]):
    queue = deque(evaluations)
    while queue:
        evaluation = queue.popleft()
        if evaluation.valid is False:
            continue
        if evaluation.evaluations:
            queue.extend(evaluation.evaluations)
        if evaluation.instance is instance and evaluation.value is not None:
            yield evaluation


class TypeValidator:
    def __init__(
        self,
        types: list[str],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.types = types
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance) -> Iterable[Evaluation]:
        valid = is_type(instance.value, *self.types)
        yield render(self, valid=valid, subject=instance)

    def __repr__(self):
        return f"TypeValidator({self.types!r})"


class Not:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.evaluator(instance)
        valid = any(e.valid is False for e in evaluations)
        yield render(self, valid=valid, subject=instance, evaluations=evaluations)


class Annotate:
    def __init__(
        self,
        value: Any,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.value = value
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        yield render(self, valid=True, value=self.value, subject=instance)


class ConditionalBase:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.evaluator(instance)
        valid = not any(e.valid is False for e in evaluations)
        yield render(self, valid=valid, subject=instance, evaluations=evaluations)


class If(ConditionalBase):
    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.evaluator(instance)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self, valid=True, value=valid, subject=instance, evaluations=evaluations
        )


class Then(ConditionalBase):
    pass


class Else(ConditionalBase):
    pass


class Conditional:
    def __init__(self, condition: If, when_true: Then = None, when_false: Else = None):
        self.condition = condition
        self.when_true = when_true
        self.when_false = when_false

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.condition(instance)
        predicate = not any(e.value is False for e in evaluations)
        if predicate and self.when_true:
            evaluations += self.when_true(instance)
        if not predicate and self.when_false:
            evaluations += self.when_false(instance)
        yield from evaluations


class Sequence:
    def __init__(
        self,
        evaluators: list[Evaluate],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluators = evaluators
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def evaluate(self, instance: Instance, /) -> Iterable[Evaluation]:
        for evaluator in self.evaluators:
            yield from evaluator(instance)


class AllOf(Sequence):
    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.evaluate(instance)
        valid = not any(e.valid is False for e in evaluations)
        yield render(self, valid=valid, subject=instance, evaluations=evaluations)


class AnyOf(Sequence):
    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.evaluate(instance)
        valid = any(e.valid is True for e in evaluations)
        yield render(self, valid=valid, subject=instance, evaluations=evaluations)


class OneOf(Sequence):
    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.evaluate(instance)
        valid = sum(e.valid is True for e in evaluations) == 1
        yield render(self, valid=valid, subject=instance, evaluations=evaluations)


class Mapping:
    def __init__(
        self,
        evaluators: dict[str, Evaluate],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluators = evaluators
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location


class Properties(Mapping):
    def __call__(self, instance: Instance[dict], /) -> Iterable[Evaluation]:
        evaluated = set()
        evaluations: list[Evaluation] = []
        for name, sub in instance:
            if name in self.evaluators:
                evaluations += self.evaluators[name](sub)
                evaluated.add(name)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )

    def __repr__(self):
        return f"Properties({self.evaluators!r})"


class PatternProperties(Mapping):
    def __call__(self, instance: Instance[dict], /) -> Iterable[Evaluation]:
        evaluated = set()
        evaluations: list[Evaluation] = []
        for name, sub in instance:
            for regex, evaluator in self.evaluators.items():
                if re_match(regex, name):
                    evaluations += evaluator(sub)
                    evaluated.add(name)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )


class AdditionalProperties:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(
        self, instance: Instance[dict], /, *, ignore: set[str] = None
    ) -> Iterable[Evaluation]:
        evaluated: set[str] = set()
        evaluations: list[Evaluation] = []
        ignore = ignore or set()
        for name, sub in instance:
            if name in ignore:
                continue
            evaluated.add(name)
            evaluations += self.evaluator(sub)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )

    def __repr__(self):
        return f"AdditionalProperties({self.evaluator!r})"


class PropertiesRules:
    def __init__(
        self,
        props: Properties = None,
        pattern_props: PatternProperties = None,
        additional_props: AdditionalProperties = None,
    ):
        self.props = props
        self.pattern_props = pattern_props
        self.additional_props = additional_props

    def __call__(self, instance: Instance[dict], /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        if evaluator := cast(Evaluate, self.props):
            evaluations += evaluator(instance)
        if evaluator := cast(Evaluate, self.pattern_props):
            evaluations += evaluator(instance)
        evaluated: set[str] = extract_evaluated(evaluations, type=int)
        if additional_props := self.additional_props:
            evaluations += additional_props(instance, ignore=evaluated)
        yield from evaluations

    def __repr__(self):
        return f"PropertiesRules(props={self.props!r}, pattern_props={self.pattern_props!r}, additional_props={self.additional_props!r})"


class PropertyNames:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[dict], /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        for name in instance.names:
            evaluations += self.evaluator(name)
        valid = not any(e.valid is False for e in evaluations)
        yield render(self, valid=valid, subject=instance, evaluations=evaluations)


class DependentSchemas(Mapping):
    def __call__(self, instance: Instance[dict], /) -> Iterable[Evaluation]:
        evaluated = set()
        evaluations: list[Evaluation] = []
        for name, _ in instance:
            if name in self.evaluators:
                evaluations += self.evaluators[name](instance)
                evaluated.add(name)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )


class MinProperties:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[dict], /) -> Iterable[Evaluation]:
        valid = self.limit <= len(instance.value)
        yield render(self, valid=valid, subject=instance)


class MaxProperties:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[dict], /) -> Iterable[Evaluation]:
        valid = self.limit >= len(instance.value)
        yield render(self, valid=valid, subject=instance)


class MinLength:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[str], /) -> Iterable[Evaluation]:
        valid = self.limit <= len(instance.value)
        yield render(self, valid=valid, subject=instance)


class MaxLength:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[str], /) -> Iterable[Evaluation]:
        valid = self.limit >= len(instance.value)
        yield render(self, valid=valid, subject=instance)


class Pattern:
    def __init__(
        self,
        regex: str,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.regex = regex
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[str], /) -> Iterable[Evaluation]:
        valid = re_match(self.regex, instance.value)
        yield render(self, valid=valid, subject=instance)


class PrefixItems:
    def __init__(
        self,
        evaluators: list[Evaluate],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluators = evaluators
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[list], /) -> Iterable[Evaluation]:
        evaluated: set[int] = set()
        evaluations: list[Evaluation] = []
        for (i, sub), evaluator in zip(instance, self.evaluators):
            evaluated.add(cast(int, i))
            evaluations += evaluator(sub)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )


class AdditionalItems:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(
        self, instance: Instance[list], /, *, ignore: set[int] = None
    ) -> Iterable[Evaluation]:
        evaluated: set[int] = set()
        evaluations: list[Evaluation] = []
        ignore = ignore or set()
        for name, sub in instance:
            if name in ignore:
                continue
            evaluated.add(name)
            evaluations += self.evaluator(sub)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )

    def __repr__(self):
        return f"AdditionalItems(evaluator={self.evaluator!r})"


class ItemsRules:
    def __init__(
        self,
        prefix_items: PrefixItems = None,
        additional_items: AdditionalItems = None,
    ):
        self.prefix_items = prefix_items
        self.additional_items = additional_items

    def __call__(self, instance: Instance[list], /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        if prefix_items := self.prefix_items:
            evaluations += prefix_items(instance)
        evaluated: set[int] = extract_evaluated(evaluations, type=int)
        if additional_items := self.additional_items:
            evaluations += additional_items(instance, ignore=evaluated)
        yield from evaluations

    def __repr__(self):
        return f"ItemsRules(prefix_items={self.prefix_items!r}, additional_items={self.additional_items!r})"


class MinItems:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[list], /) -> Iterable[Evaluation]:
        valid = self.limit <= len(instance.value)
        yield render(self, valid=valid, subject=instance)

    def __repr__(self):
        return f"MinItems({self.limit!r})"


class MaxItems:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[list], /) -> Iterable[Evaluation]:
        valid = self.limit >= len(instance.value)
        yield render(self, valid=valid, subject=instance)

    def __repr__(self):
        return f"MaxItems({self.limit!r})"


class UniqueItems:
    def __init__(
        self,
        activated: bool,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.activated = activated
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[list], /) -> Iterable[Evaluation]:
        if self.activated:
            found = []
            valid = True
            for element in instance.value:
                e = make_comparable(element)
                if e in found:
                    valid = False
                    break
                found.append(e)
        yield render(self, valid=valid, subject=instance)


class Contains:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[list], /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluated = set()
        for index, sub in instance:
            tmp = list(self.evaluator(sub))
            if all(t.valid is True for t in tmp):
                evaluated.add(index)
            evaluations += tmp
        valid = bool(evaluated)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )


class MinContains:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[set[int]], /) -> Iterable[Evaluation]:
        valid = self.limit <= len(instance.value)
        yield render(self, valid=valid, subject=instance)


class MaxContains:
    def __init__(
        self,
        limit: int,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: InstanceContext[set[int]], /) -> Iterable[Evaluation]:
        valid = self.limit >= len(instance.value)
        yield render(self, valid=valid, subject=instance)


class ContainmentRules:
    def __init__(
        self,
        contains: Contains,
        minimum: MinContains = None,
        maximum: MaxContains = None,
    ):
        self.contains = contains
        self.minimum = minimum
        self.maximum = maximum

    def __call__(self, instance: Instance[list], /) -> Iterable[Evaluation]:
        evaluations: list[Evaluation] = []
        evaluations += self.contains(instance)
        yield from evaluations
        evaluated = set()
        for evaluation in evaluations:
            if evaluation.valid is True:
                evaluated.update(evaluation.value)
        if self.minimum:
            yield from self.minimum(instance.ctx(evaluated))
        if self.maximum:
            yield from self.maximum(instance.ctx(evaluated))


class UnevaluatedProperties:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(
        self, instance: Instance[dict], /, *, ignore: set[str] = None
    ) -> Iterable[Evaluation]:
        evaluated: set[str] = set()
        evaluations: list[Evaluation] = []
        ignore = ignore or set()
        for name, sub in instance:
            if name in ignore:
                continue
            evaluated.add(name)
            evaluations += self.evaluator(sub)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )


class UnevaluatedItems:
    def __init__(
        self,
        evaluator: Evaluate,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.evaluator = evaluator
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(
        self, instance: Instance[list], /, *, ignore: set[int] = None
    ) -> Iterable[Evaluation]:
        evaluated: set[int] = set()
        evaluations: list[Evaluation] = []
        ignore = ignore or set()
        for name, sub in instance:
            if name in ignore:
                continue
            evaluated.add(name)
            evaluations += self.evaluator(sub)
        valid = not any(e.valid is False for e in evaluations)
        yield render(
            self,
            valid=valid,
            subject=instance,
            value=evaluated,
            evaluations=evaluations,
        )


class Enum:
    def __init__(
        self,
        values: list[Any],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.allowed_values = values
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance) -> Iterable[Evaluation]:
        comparable = make_comparable(instance.value)
        valid = any(comparable == make_comparable(val) for val in self.allowed_values)
        yield render(self, valid=valid, subject=instance)


class Const:
    def __init__(
        self,
        value: Any,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.allowed_value = value
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance) -> Iterable[Evaluation]:
        valid = make_comparable(instance.value) == make_comparable(self.allowed_value)
        yield render(self, valid=valid, subject=instance)


class Required:
    def __init__(
        self,
        requirements: list[str],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.requirements = requirements
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[dict]) -> Iterable[Evaluation]:
        missing = set(self.requirements) - set(instance.value)
        valid = not missing
        yield render(self, valid=valid, subject=instance)

    def __repr__(self):
        return f"Required({self.requirements!r})"


class DependentRequired:
    def __init__(
        self,
        dependencies: dict[str, list[str]],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.dependencies = dependencies
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[dict]) -> Iterable[Evaluation]:
        missing = set()
        for key, requirements in self.dependencies.items():
            if key in instance.value:
                missing |= set(requirements) - set(instance.value)
        valid = not missing
        yield render(self, valid=valid, subject=instance)


class Maximum:
    def __init__(
        self,
        limit: float,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[float]) -> Iterable[Evaluation]:
        valid = self.limit >= instance.value
        yield render(self, valid=valid, subject=instance)


class ExclusiveMaximum:
    def __init__(
        self,
        limit: float,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[float]) -> Iterable[Evaluation]:
        valid = self.limit > instance.value
        yield render(self, valid=valid, subject=instance)


class Minimum:
    def __init__(
        self,
        limit: float,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[float]) -> Iterable[Evaluation]:
        valid = self.limit <= instance.value
        yield render(self, valid=valid, subject=instance)


class ExclusiveMinimum:
    def __init__(
        self,
        limit: float,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.limit = limit
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[float]) -> Iterable[Evaluation]:
        valid = self.limit < instance.value
        yield render(self, valid=valid, subject=instance)


class MultipleOf:
    def __init__(
        self,
        operand: float,
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.operand = operand
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[float]) -> Iterable[Evaluation]:
        if self.operand > instance.value:
            yield render(
                self,
                valid=False,
                subject=instance,
                reason=f"Value is not a multiple of {self.operand}",
            )
        elif instance.value == self.operand:
            yield render(self, valid=True, subject=instance)
        elif instance.value % self.operand == 0:
            yield render(self, valid=True, subject=instance)
        else:
            yield render(
                self,
                valid=False,
                subject=instance,
                reason=f"Value is not a multiple of {self.operand}",
            )


class Format:
    def __init__(
        self,
        checker: Callable[[Any], bool | NotImplementedType],
        relative_location: str,
        absolute_keyword_location: str,
    ):
        self.checker = checker
        self.relative_location = relative_location
        self.absolute_keyword_location = absolute_keyword_location

    def __call__(self, instance: Instance[float]) -> Iterable[Evaluation]:
        if (valid := self.checker(instance.value)) is not NotImplemented:
            yield render(self, valid=valid, subject=instance)


class Only:
    def __init__(
        self,
        evaluators: list[Evaluate],
        type: str,
    ):
        self.evaluators = evaluators
        self.type = type

    def __call__(self, instance: Instance, /) -> Iterable[Evaluation]:
        if is_type(instance.value, self.type):
            for evaluator in self.evaluators:
                yield from evaluator(instance)

    def __repr__(self):
        return f"Only(type={self.type!r}, evaluators={self.evaluators!r})"


def extract_evaluated(evaluations: list[Evaluation], type=Type[T]) -> set[T]:
    return reduce(
        lambda a, b: a | b,
        (c.value for c in evaluations if c.valid is True),
        set(),
    )


def re_match(regex, value):
    import re

    return bool(re.search(regex, value))


from functools import singledispatch

BOOLEAN_TRUE = object()
BOOLEAN_FALSE = object()


@singledispatch
def make_comparable(obj):
    return obj


@make_comparable.register(bool)
def make_comparable_bool(obj):
    if obj is True:
        return BOOLEAN_TRUE
    if obj is False:
        return BOOLEAN_FALSE


@make_comparable.register(float)
def make_comparable_float(obj: float):
    return Decimal(str(obj))


@make_comparable.register(int)
def make_comparable_int(obj: int):
    return Decimal(obj)


@make_comparable.register(dict)
def make_comparable_dict(obj):
    return {key: make_comparable(val) for key, val in obj.items()}


@make_comparable.register(list)
def make_comparable_list(obj):
    return [make_comparable(element) for element in obj]



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Deque, Generic, TypeAlias, TypeVar, Union

from .ref import Ref

if TYPE_CHECKING:
    from .reg import Registry

DEFAULT_DIALECT = "https://json-schema.org/draft/2020-12/schema"

DictSchema: TypeAlias = dict
BoolSchema: TypeAlias = bool
Schema = Union[DictSchema, BoolSchema]

T = TypeVar("T")
S = TypeVar("S", bound=Schema)


@dataclass(repr=False)
class Document(Generic[S]):
    schema: S
    dialect: str
    canonical_uri: str | None = None
    anchor: str | None = None
    relative_location: str = ""
    root: Document | None = None
    parent: Document | None = None
    ref: Ref | None = None
    ref_doc: Document | None = None

    def __post_init__(self):
        if self.relative_location and not self.root:
            raise Exception("relative location and root are mandatory")
        if not self.relative_location and self.root:
            raise Exception("relative location and root are mandatory")

    def __repr__(self):
        return (
            f"Document("
            f"canonical_uri={self.canonical_uri!r}, "
            f"anchor={self.anchor!r}, "
            f"relative_location={self.relative_location!r}, "
            f"ref={self.ref!r})"
        )

    @property
    def absolute_location(self):
        if self.canonical_uri:
            return self.canonical_uri
        if self.root:
            return (self.root.canonical_uri or "") + "#" + self.relative_location


def document_factory(
    schema: S,
    *,
    relative_location: str = "",
    root: Document = None,
    parent: Document = None,
    reg: Registry,
    default_dialect: str = DEFAULT_DIALECT,
) -> Document[S]:
    dialect = guess_dialect(schema, parent, default_dialect)
    doc = Document(
        schema=schema,
        relative_location=relative_location,
        root=root,
        parent=parent,
        dialect=dialect,
    )
    reg.finalize(doc)
    return doc


def guess_dialect(schema: dict | bool, parent: Document | None, default: str) -> str:
    if isinstance(schema, dict) and (dialect := schema.get("$schema")):
        return dialect
    elif parent and (dialect := parent.dialect):
        return dialect
    return default


def canonical_uri_join(base: str, canonical_uri: str) -> str:
    from urllib.parse import urljoin

    if base.startswith(("http://", "https://")):
        return urljoin(base, canonical_uri, allow_fragments=False)
    return canonical_uri


def canonical_ref_join(base: str, ref: str) -> str:
    from urllib.parse import urljoin

    if base.startswith(("http://", "https://")):
        base, *_ = base.partition("#")
        ref = urljoin(base, ref, allow_fragments=True)
    if "#" not in ref:
        ref = ref + "#"
    return ref


def resolve(schema: S, *, reg: Registry, default_dialect: str = None) -> Document[S]:
    default_dialect = default_dialect or DEFAULT_DIALECT
    doc = document_factory(schema, default_dialect=default_dialect, reg=reg)
    resolve_2(doc, reg)
    return doc


def resolve_2(doc: Document, reg: Registry):
    from collections import deque

    queue1: Deque[Document] = deque([doc])  # explore down
    queue2: Deque[Document] = deque([])  # check for $refs
    while queue1 or queue2:
        while queue1:
            doc = queue1.popleft()
            reg.documents.append(doc)
            for schema, relative_location in reg.extract(doc):
                queue1.append(
                    document_factory(
                        schema=schema,
                        relative_location=relative_location,
                        root=doc.root or doc,
                        parent=doc,
                        reg=reg,
                    )
                )
            queue2.append(doc)
        while queue2:
            doc = queue2.popleft()
            if doc.ref:
                if ref_doc := find_by_ref(reg, doc):
                    doc.ref_doc = ref_doc
                else:
                    schema, *_ = reg.load(doc.ref)
                    queue2.append(doc)
                    doc = document_factory(schema, reg=reg)
                    queue1.append(doc)
                    # doc.ref_doc = reg.load(ref)
                    raise NotImplementedError(
                        "Need to grab from outher space", reg.documents
                    )

            # raise NotImplementedError
        continue


def find_by_ref(reg: Registry, doc: Document):
    if doc.ref is None:
        raise
    uri, pointer, anchor = doc.ref.decomposed
    relative_location = pointer or ""
    if uri:
        documents = list(d for d in reg.documents if d.canonical_uri == uri)
    elif root := doc.root:
        documents = [d for d in reg.documents if d.root == root] + [root]
    elif doc.relative_location == "":
        documents = [d for d in reg.documents if d.root == doc] + [doc]
    else:
        raise
    if anchor:
        for document in reg.documents:
            if document.anchor == anchor:
                return document
    else:
        for document in documents:
            if document.relative_location == relative_location:
                return document



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Callable, Iterator

from .ref import push_fragments

if TYPE_CHECKING:
    from types import NotImplementedType

    from .evaluators import Evaluate
    from .reg import Registry
    from .res import Document

    relative_location: str
    formats: dict[str, Callable[[Any], bool | NotImplementedType]]

reg_var: ContextVar[Registry] = ContextVar("reg_var")
relative_location_var: ContextVar[str] = ContextVar("relative_location_var", default="")


def find(doc: Document, key: str) -> Document | None:
    return reg_var.get().find(doc, key)


def find_sequence(doc: Document, key: str) -> list[tuple[int, Document]] | None:
    return reg_var.get().find_sequence(doc, key)


def find_mapping(doc: Document, key: str) -> list[tuple[str, Document]] | None:
    return reg_var.get().find_mapping(doc, key)


def compile_schema(doc: Document) -> Evaluate:
    return reg_var.get().compile_schema(doc)


@contextmanager
def push_relative_location(*fragments: int | str) -> Iterator[str]:
    relative_location = push_fragments(relative_location_var.get(), *fragments)
    try:
        token = relative_location_var.set(relative_location)
        yield relative_location
    finally:
        relative_location_var.reset(token)


@contextmanager
def push_registry(reg: Registry):
    try:
        token = reg_var.set(reg)
        yield
    finally:
        reg_var.reset(token)


def __getattr__(name):
    if name == "relative_location":
        return relative_location_var.get()
    if name == "formats":
        return reg_var.get().formats
    raise AttributeError



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING, Iterable, Type

from ilia import ctx
from ilia.checkers import (
    is_date,
    is_date_time,
    is_duration,
    is_email,
    is_hostname,
    is_idn_email,
    is_idn_hostname,
    is_ipv4,
    is_ipv6,
    is_iri,
    is_iri_reference,
    is_json_pointer,
    is_regex,
    is_relative_json_pointer,
    is_time,
    is_uri,
    is_uri_reference,
    is_uri_template,
    is_uuid,
)
from ilia.evaluators import (
    AdditionalItems,
    AdditionalProperties,
    AllOf,
    AlwaysFalse,
    AlwaysTrue,
    Annotate,
    AnyOf,
    Cluster,
    Conditional,
    Const,
    ContainmentRules,
    Contains,
    DependentRequired,
    DependentSchemas,
    Else,
    Enum,
    ExclusiveMaximum,
    ExclusiveMinimum,
    Format,
    If,
    ItemsRules,
    MaxContains,
    Maximum,
    MaxItems,
    MaxLength,
    MaxProperties,
    MinContains,
    Minimum,
    MinItems,
    MinLength,
    MinProperties,
    MultipleOf,
    Not,
    OneOf,
    Only,
    Pattern,
    PatternProperties,
    PrefixItems,
    Properties,
    PropertiesRules,
    PropertyNames,
    Required,
    Then,
    TypeValidator,
    UnevaluatedItems,
    UnevaluatedProperties,
    UniqueItems,
)
from ilia.ref import Ref, push_fragments, push_pointer

if TYPE_CHECKING:
    from ilia.evaluators import Evaluate
    from ilia.res import Document, Schema

DIALECT = "https://json-schema.org/draft/2020-12/schema"
FORMATS = {
    "date-time": is_date_time,
    "date": is_date,
    "time": is_time,
    "duration": is_duration,
    "email": is_email,
    "idn-email": is_idn_email,
    "hostname": is_hostname,
    "idn-hostname": is_idn_hostname,
    "ipv4": is_ipv4,
    "ipv6": is_ipv6,
    "uri": is_uri,
    "uri-reference": is_uri_reference,
    "iri": is_iri,
    "iri-reference": is_iri_reference,
    "uuid": is_uuid,
    "uri-template": is_uri_template,
    "json-pointer": is_json_pointer,
    "relative-json-pointer": is_relative_json_pointer,
    "regex": is_regex,
}
Missing = object()


def finalize(doc: Document[dict]):
    from ilia.res import canonical_ref_join, canonical_uri_join

    canonical_uri = None
    anchor = None
    ref = None
    if isinstance(doc.schema, dict):
        canonical_uri = doc.schema.get("$id")
        anchor = doc.schema.get("$anchor")
        ref = doc.schema.get("$ref")
    if canonical_uri and doc.parent and (base := doc.parent.canonical_uri):
        canonical_uri = canonical_uri_join(base, canonical_uri)
    if ref and canonical_uri:
        ref = canonical_ref_join(canonical_uri, ref)
    doc.canonical_uri = canonical_uri
    doc.anchor = anchor
    if ref:
        doc.ref = Ref(ref)


def extract(doc: Document[Schema]) -> Iterable[tuple[Schema, str]]:
    if isinstance(doc.schema, bool):
        return
    for keyword in [
        "not",
        "if",
        "else",
        "then",
        "items",
        "contains",
        "additionalProperties",
        "propertyNames",
        "unevaluatedItems",
        "unevaluatedProperties",
    ]:
        if (schema := doc.schema.get(keyword, Missing)) is not Missing:
            yield schema, push_fragments(doc.relative_location, keyword)
    for keyword in ["allOf", "anyOf", "oneOf", "prefixItems"]:
        for i, schema in enumerate(doc.schema.get(keyword, [])):
            yield schema, push_fragments(doc.relative_location, keyword, i)
    for keyword in ["$defs", "dependentSchemas", "properties", "patternProperties"]:
        for key, schema in doc.schema.get(keyword, {}).items():
            yield schema, push_fragments(doc.relative_location, keyword, key)


def compile_schema(doc: Document[dict | bool]) -> Evaluate:
    if doc.schema is True:
        return AlwaysTrue(
            relative_location=ctx.relative_location,
            absolute_keyword_location=doc.absolute_location,
        )
    elif doc.schema is False:
        return AlwaysFalse(
            relative_location=ctx.relative_location,
            absolute_keyword_location=doc.absolute_location,
        )
    else:
        keywords = list(doc.schema)

        unevaluated_items = find_unevaluated_items(doc, keywords)
        unevaluated_props = find_unevaluated_properties(doc, keywords)

        evaluators: list[Evaluate] = []
        evaluators += yield_ref(doc, keywords)
        evaluators += yield_not(doc, keywords)
        evaluators += compile_conditional(doc, keywords)
        evaluators += yield_all_of(doc, keywords)
        evaluators += yield_any_of(doc, keywords)
        evaluators += yield_one_of(doc, keywords)
        evaluators += yield_object(doc, keywords)
        evaluators += yield_array(doc, keywords)
        evaluators += yield_string(doc, keywords)
        evaluators += yield_number(doc, keywords)
        evaluators += yield_type(doc, keywords)
        evaluators += yield_enum(doc, keywords)
        evaluators += yield_const(doc, keywords)
        evaluators += yield_format(doc, keywords)
        evaluators += yield_annotations(doc, keywords)

        return Cluster(
            evaluators=evaluators,
            unevaluated_items=unevaluated_items,
            unevaluated_props=unevaluated_props,
            relative_location=ctx.relative_location,
            absolute_keyword_location=doc.absolute_location,
        )


def find_simple_evaluator(
    doc: Document, keywords: list[str], attr: str, cls: Type[Evaluate]
):
    if sub := ctx.find(doc, attr):
        with ctx.push_relative_location(attr):
            keywords.remove(attr)
            return cls(
                ctx.compile_schema(sub),
                relative_location=ctx.relative_location,
                absolute_keyword_location=sub.absolute_location,
            )


def yield_simple_evaluator(
    doc: Document, keywords: list[str], attr: str, cls: Type[Evaluate]
):
    if evaluator := find_simple_evaluator(doc, keywords, attr, cls):
        yield evaluator


def find_simple_validator(
    doc: Document, keywords: list[str], attr: str, cls: Type[Evaluate]
):
    if (limit := doc.schema.get(attr, Missing)) is not Missing:
        with ctx.push_relative_location(attr):
            keywords.remove(attr)
            return cls(
                limit,
                relative_location=ctx.relative_location,
                absolute_keyword_location=push_pointer(doc.absolute_location, attr),
            )


def yield_simple_validator(
    doc: Document, keywords: list[str], attr: str, cls: Type[Evaluate]
):
    if evaluator := find_simple_validator(doc, keywords, attr=attr, cls=cls):
        yield evaluator


yield_not = partial(yield_simple_evaluator, attr="not", cls=Not)


def compile_conditional(doc: Document, keywords: list[str]):
    when_true = find_then(doc, keywords=keywords)
    when_false = find_else(doc, keywords=keywords)
    if condition := find_if(doc, keywords=keywords):
        yield Conditional(
            condition=condition, when_false=when_false, when_true=when_true
        )


find_if = partial(find_simple_evaluator, attr="if", cls=If)


find_then = partial(find_simple_evaluator, attr="then", cls=Then)


find_else = partial(find_simple_evaluator, attr="else", cls=Else)


@ctx.push_relative_location("$ref")
def yield_ref(doc: Document, keywords: list[str]):
    if sub := doc.ref_doc:
        keywords.remove("$ref")
        yield ctx.compile_schema(sub)


@ctx.push_relative_location("anyOf")
def yield_any_of(doc: Document, keywords: list[str]):
    if docs := ctx.find_sequence(doc, "anyOf"):
        keywords.remove("anyOf")

        yield AnyOf(
            compile_sequence(docs),
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "anyOf"),
        )


@ctx.push_relative_location("allOf")
def yield_all_of(doc: Document, keywords: list[str]):
    if docs := ctx.find_sequence(doc, "allOf"):
        keywords.remove("allOf")

        yield AllOf(
            compile_sequence(docs),
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "allOf"),
        )


@ctx.push_relative_location("oneOf")
def yield_one_of(doc: Document, keywords: list[str]):
    if docs := ctx.find_sequence(doc, "oneOf"):
        keywords.remove("oneOf")

        yield OneOf(
            compile_sequence(docs),
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "oneOf"),
        )


def yield_annotations(doc: Document, keywords: list[str]):
    for keyword in keywords:
        value = doc.schema.get(keyword)
        yield Annotate(
            value=value,
            relative_location=push_fragments(ctx.relative_location, keyword),
            absolute_keyword_location=push_pointer(doc.absolute_location, keyword),
        )
    keywords.clear()


def yield_object(doc: Document, keywords: list[str]):
    evaluators: list[Evaluate] = []
    evaluators += yield_properties_rules(doc, keywords)
    evaluators += yield_property_names(doc, keywords)
    evaluators += yield_dependent_schemas(doc, keywords)
    evaluators += yield_min_properties(doc, keywords)
    evaluators += yield_max_properties(doc, keywords)
    evaluators += yield_required(doc, keywords)
    evaluators += yield_dependent_required(doc, keywords)
    if evaluators:
        yield Only(evaluators=evaluators, type="object")


def yield_properties_rules(doc: Document, keywords: list[str]):
    props = find_props(doc, keywords)
    pattern_props = find_pattern_props(doc, keywords)
    additional_props = find_additional_props(doc, keywords)
    if props or pattern_props or additional_props:
        yield PropertiesRules(
            props=props, pattern_props=pattern_props, additional_props=additional_props
        )


def compile_sequence(docs: list[tuple[int, Document]]) -> list[Evaluate]:
    evaluators = []
    for index, sub in docs:
        with ctx.push_relative_location(index):
            evaluators.append(ctx.compile_schema(sub))
    return evaluators


def compile_mapping(docs: list[tuple[str, Document]]) -> dict[str, Evaluate]:
    evaluators = {}
    for name, sub in docs:
        with ctx.push_relative_location(name):
            evaluators[name] = ctx.compile_schema(sub)
    return evaluators


@ctx.push_relative_location("properties")
def find_props(doc: Document, keywords: list[str]):
    if docs := ctx.find_mapping(doc, "properties"):
        keywords.remove("properties")
        return Properties(
            compile_mapping(docs),
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "properties"),
        )


@ctx.push_relative_location("patternProperties")
def find_pattern_props(doc: Document, keywords: list[str]):
    if docs := ctx.find_mapping(doc, "patternProperties"):
        keywords.remove("patternProperties")
        return PatternProperties(
            compile_mapping(docs),
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(
                doc.absolute_location, "patternProperties"
            ),
        )


find_additional_props = partial(
    find_simple_evaluator, attr="additionalProperties", cls=AdditionalProperties
)


yield_property_names = partial(
    yield_simple_evaluator, attr="propertyNames", cls=PropertyNames
)


@ctx.push_relative_location("dependentSchemas")
def yield_dependent_schemas(doc: Document, keywords: list[str]):
    if docs := ctx.find_mapping(doc, "dependentSchemas"):
        keywords.remove("dependentSchemas")
        yield DependentSchemas(
            compile_mapping(docs),
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(
                doc.absolute_location, "dependentSchemas"
            ),
        )


yield_min_properties = partial(
    yield_simple_validator, attr="minProperties", cls=MinProperties
)


yield_max_properties = partial(
    yield_simple_validator, attr="maxProperties", cls=MaxProperties
)


yield_required = partial(yield_simple_validator, attr="required", cls=Required)


yield_dependent_required = partial(
    yield_simple_validator, attr="dependentRequired", cls=DependentRequired
)


def yield_array(doc: Document, keywords: list[str]):
    evaluators: list[Evaluate] = []
    evaluators += yield_items_rules(doc, keywords)
    evaluators += yield_containment(doc, keywords)
    evaluators += yield_max_items(doc, keywords)
    evaluators += yield_min_items(doc, keywords)
    evaluators += yield_unique_items(doc, keywords)
    if evaluators:
        yield Only(evaluators=evaluators, type="array")


def yield_items_rules(doc: Document, keywords: list[str]):
    prefix_items = find_prefix_items(doc, keywords)
    additional_items = find_additional_items(doc, keywords)
    if prefix_items or additional_items:
        yield ItemsRules(prefix_items=prefix_items, additional_items=additional_items)


@ctx.push_relative_location("prefixItems")
def find_prefix_items(doc: Document, keywords: list[str]):
    if docs := ctx.find_sequence(doc, "prefixItems"):
        keywords.remove("prefixItems")
        evaluators: list[Evaluate] = []
        for index, sub in docs:
            with ctx.push_relative_location(index):
                evaluators.append(ctx.compile_schema(sub))
        return PrefixItems(
            evaluators,
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(
                doc.absolute_location, "prefixItems"
            ),
        )


find_additional_items = partial(
    find_simple_evaluator, attr="items", cls=AdditionalItems
)

yield_min_items = partial(yield_simple_validator, attr="minItems", cls=MinItems)


yield_max_items = partial(yield_simple_validator, attr="maxItems", cls=MaxItems)
yield_unique_items = partial(
    yield_simple_validator, attr="uniqueItems", cls=UniqueItems
)


def yield_containment(doc: Document, keywords: list[str]):
    maximum = find_max_contains(doc, keywords)
    minimum = find_min_contains(doc, keywords)
    if contains := find_contains(doc, keywords):
        yield ContainmentRules(contains=contains, minimum=minimum, maximum=maximum)


find_contains = partial(find_simple_evaluator, attr="contains", cls=Contains)


find_max_contains = partial(find_simple_validator, attr="maxContains", cls=MaxContains)


find_min_contains = partial(find_simple_validator, attr="minContains", cls=MinContains)


def yield_string(doc: Document, keywords: list[str]):
    evaluators: list[Evaluate] = []
    evaluators += yield_pattern(doc, keywords)
    evaluators += yield_min_length(doc, keywords)
    evaluators += yield_max_length(doc, keywords)
    if evaluators:
        yield Only(evaluators=evaluators, type="string")


yield_pattern = partial(yield_simple_validator, attr="pattern", cls=Pattern)


yield_min_length = partial(yield_simple_validator, attr="minLength", cls=MinLength)


yield_max_length = partial(yield_simple_validator, attr="maxLength", cls=MaxLength)


def yield_number(doc: Document, keywords: list[str]):
    evaluators: list[Evaluate] = []
    evaluators += yield_minimum(doc, keywords)
    evaluators += yield_maximum(doc, keywords)
    evaluators += yield_exclusive_minimum(doc, keywords)
    evaluators += yield_exclusive_maximum(doc, keywords)
    evaluators += yield_multiple_of(doc, keywords)
    if evaluators:
        yield Only(evaluators=evaluators, type="number")


yield_multiple_of = partial(yield_simple_validator, attr="multipleOf", cls=MultipleOf)


yield_minimum = partial(yield_simple_validator, attr="minimum", cls=Minimum)


yield_maximum = partial(yield_simple_validator, attr="maximum", cls=Maximum)


yield_exclusive_minimum = partial(
    yield_simple_validator, attr="exclusiveMinimum", cls=ExclusiveMinimum
)


yield_exclusive_maximum = partial(
    yield_simple_validator, attr="exclusiveMaximum", cls=ExclusiveMaximum
)


@ctx.push_relative_location("type")
def yield_type(doc: Document, keywords: list[str]):
    if (value := doc.schema.get("type", Missing)) is not Missing:
        keywords.remove("type")
        yield TypeValidator(
            value if isinstance(value, list) else [value],
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "type"),
        )


@ctx.push_relative_location("enum")
def yield_enum(doc: Document, keywords: list[str]):
    if (values := doc.schema.get("enum", Missing)) is not Missing:
        keywords.remove("enum")
        yield Enum(
            values,
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "enum"),
        )


@ctx.push_relative_location("const")
def yield_const(doc: Document, keywords: list[str]):
    if (value := doc.schema.get("const", Missing)) is not Missing:
        keywords.remove("const")
        yield Const(
            value,
            relative_location=ctx.relative_location,
            absolute_keyword_location=push_pointer(doc.absolute_location, "const"),
        )


def yield_format(doc: Document, keywords: list[str]):
    if (value := doc.schema.get("format", Missing)) is not Missing:
        with ctx.push_relative_location("format"):
            keywords.remove("format")
            checker = ctx.formats.get(value) or FORMATS[value]
            yield Format(
                checker,
                relative_location=ctx.relative_location,
                absolute_keyword_location=push_pointer(doc.absolute_location, "format"),
            )


find_unevaluated_items = partial(
    find_simple_evaluator, attr="unevaluatedItems", cls=UnevaluatedItems
)


find_unevaluated_properties = partial(
    find_simple_evaluator, attr="unevaluatedProperties", cls=UnevaluatedProperties
)



from __future__ import annotations

from .reg import Registry, registry_factory
from .res import Document, Schema, resolve
from .validators import Validator
from . import ctx

__all__ = [
    "Document",
    "Registry",
    "registry_factory",
    "resolve",
]


def load(schema: Schema, reg: Registry = None):
    reg = reg or registry_factory()
    with ctx.push_registry(reg):
        doc = resolve(schema, reg=reg)
        evaluator = reg.compile_schema(doc)
        return Validator(doc, evaluator)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['ilia', 'ilia.dialects']

package_data = \
{'': ['*']}

install_requires = \
['httpx[brotli,http2]>=0.20.0,<0.21.0',
 'pytest-subtests>=0.5.0,<0.6.0',
 'types-dataclasses>=0.6.1,<0.7.0']

setup_kwargs = {
    'name': 'ilia',
    'version': '0.1.0',
    'description': 'a JSON Schema draft 2020-12 validation library',
    'long_description': '# ilia\n',
    'author': 'Xavier Barbosa',
    'author_email': 'clint.northwood@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://lab.errorist.xyz',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    from .backends import Backend
    from .res import Schema


class Loader:
    def __init__(self, backends: list[Backend]):
        self.backends = backends

    def get(self, uri: str):
        for backend in self.backends:
            if backend.applicable(uri):
                return backend.get(uri)
        raise



from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from .evaluators import Instance

if TYPE_CHECKING:
    from .evaluators import Evaluate, Evaluation
    from .res import Document


@dataclass
class Validation:
    valid: bool
    instance_value: Any
    evaluations: list[Evaluation]


class Validator:
    def __init__(self, doc: Document, evaluator: Evaluate):
        self.doc = doc
        self.evaluator = evaluator

    def validate(self, value: Any, /) -> Validation:
        instance = Instance("", value=value)
        evaluations: list[Evaluation] = []
        evaluations += self.evaluator(instance)
        valid = not any(e.valid is False for e in evaluations)
        return Validation(valid, instance_value=value, evaluations=evaluations)
